import ast
import time
from datetime import datetime

from pandas import DataFrame, merge, concat
from sklearn.preprocessing import MinMaxScaler

from offline_results.common.config import (
    CONFIG_HOMEPAGE_PAYTV,
    CONFIG_HOMEPAGE_NO_PAYTV,
)
from offline_results.common.constants import (
    CONTENT_ID,
    VIEW_COUNT,
    DURATION,
    SCALED_DURATION,
    SCALED_VIEW_COUNT,
    INNER,
    UBD_CREATED_ON,
    SCORE,
    CREATED_ON,
    LEFT,
    HOMEPAGE_ID,
    SERVICE_NAME,
    SEASONAL_MODULE_NAME,
    RECORDS,
    HOMEPAGE_ID_BASED,
    PAY_TV,
    NO_PAY_TV,
    REC_TYPE,
    IS_PAY_TV,
    MODEL_NAME,
    VIEW_HISTORY,
    RECOMMENDATION_SCORE,
    RANK_SCORE, ALL_CONTENT_BASED,
)
from offline_results.recommendation.homepage_recommendation.trending.trending_model import (
    TrendingModel,
)
from offline_results.recommendation.utils import RecommendationUtils
from offline_results.repository.graph_db_connection import ANGraphDb
from offline_results.utils import custom_exception, Logging


class SeasonalModel:
    viewed_relation_history_df = DataFrame()
    live_tv_channel_paytv = []
    live_tv_channel_nopaytv = []

    @staticmethod
    @custom_exception()
    def get_data():
        obj = SeasonalModel
        if len(obj.viewed_relation_history_df) != 0:
            return
        obj.viewed_relation_history_df = RecommendationUtils.user_viewed_data_from_s3()
        obj.live_tv_channel_paytv = RecommendationUtils.get_tv_channels(is_pay_tv=True)
        obj.live_tv_channel_nopaytv = RecommendationUtils.get_tv_channels(is_pay_tv=False)

    @staticmethod
    @custom_exception()
    def clear_data():
        obj = SeasonalModel
        obj.viewed_relation_history_df = DataFrame()
        obj.live_tv_channel_paytv = []
        obj.live_tv_channel_nopaytv = []

    @staticmethod
    @custom_exception()
    def get_dict_format_output(df, key_prefix, homepage_id_wise):
        if homepage_id_wise:
            output_dict = {}
            unique_homepage_id = df[HOMEPAGE_ID].unique()
            for homepage_id in unique_homepage_id:
                key_prefix_cls_hid = key_prefix + ":" + str(homepage_id)
                homepage_wise_df = df.loc[df[HOMEPAGE_ID] == homepage_id]
                homepage_wise_df = homepage_wise_df[
                    [CONTENT_ID, CREATED_ON, REC_TYPE, SCORE]
                ]
                output_dict[key_prefix_cls_hid] = homepage_wise_df.to_dict(RECORDS)
        else:
            df = df[[CONTENT_ID, HOMEPAGE_ID, CREATED_ON, REC_TYPE, SCORE]]
            output_dict = {key_prefix: df.to_dict(RECORDS)}
        return output_dict

    @staticmethod
    @custom_exception()
    def filter_by_dedicated_homepage_id(df, user_label):
        config_dictionary = (
            CONFIG_HOMEPAGE_PAYTV if user_label == PAY_TV else CONFIG_HOMEPAGE_NO_PAYTV
        )
        config_df = DataFrame(config_dictionary.items())
        config_df.columns = [HOMEPAGE_ID, MODEL_NAME]
        config_df = config_df[config_df[MODEL_NAME] == SEASONAL_MODULE_NAME.lower()]
        list_dedicated_homepage_id = config_df[HOMEPAGE_ID].tolist()
        df = df.explode(column=HOMEPAGE_ID)
        df = df[df[HOMEPAGE_ID].isin(list_dedicated_homepage_id)].reset_index(drop=True)
        return df

    @staticmethod
    @custom_exception()
    def fetch_historical_data(graph, user_label, homepage_id_wise):
        try:
            Logging.info("Fetch content-homepage_id mapping")
            (
                content_homepage_id_mapping,
                content_homepage_df,
            ) = TrendingModel.get_content_homepage_id_mapping(
                graph, user_label, homepage_id_wise
            )
        except Exception:
            graph = ANGraphDb.new_connection_config().graph
            Logging.info("Re-trying to Fetch content-homepage_id mapping")
            (
                content_homepage_id_mapping,
                content_homepage_df,
            ) = TrendingModel.get_content_homepage_id_mapping(
                graph, user_label, homepage_id_wise
            )
        Logging.info("Load Viewed Relation Network from S3")
        viewed_relation_history_df = SeasonalModel.viewed_relation_history_df
        is_pay_tv_status = True if user_label == PAY_TV else False
        viewed_relation_history_df = viewed_relation_history_df[
            viewed_relation_history_df[IS_PAY_TV] == is_pay_tv_status
            ].reset_index(drop=True)
        if not homepage_id_wise:
            viewed_relation_history_df = viewed_relation_history_df[
                ~viewed_relation_history_df[CONTENT_ID].isin(
                    SeasonalModel.live_tv_channel_paytv if user_label == PAY_TV else SeasonalModel.live_tv_channel_nopaytv
                )
            ].reset_index(drop=True)

        Logging.info("Generating UBD Dataframe for " + user_label + " users")
        view_history = viewed_relation_history_df[VIEW_HISTORY].to_list()
        view_history2 = [ast.literal_eval(str(i)) for i in view_history]
        history_content_df = DataFrame(
            {
                CONTENT_ID: viewed_relation_history_df[CONTENT_ID].to_list(),
                VIEW_HISTORY: view_history2,
            }
        )
        history_content_df = history_content_df.explode(column=VIEW_HISTORY)
        history_content_df[DURATION] = history_content_df[VIEW_HISTORY].apply(
            lambda x: x[DURATION]
        )
        history_content_df[UBD_CREATED_ON] = history_content_df[VIEW_HISTORY].apply(
            lambda x: x[CREATED_ON]
        )
        history_content_df = merge(
            history_content_df, content_homepage_id_mapping, on=CONTENT_ID, how=LEFT
        )
        Logging.info("Filter by dedicated homepage_id for Seasonal Model")
        history_content_df = SeasonalModel.filter_by_dedicated_homepage_id(
            history_content_df, user_label
        )
        return history_content_df, content_homepage_id_mapping

    @staticmethod
    @custom_exception()
    def homepage_id_wise_scaling(history_content_df):
        seasonal_content_df = DataFrame()
        homepage_ids = history_content_df[HOMEPAGE_ID].unique()
        for idx in homepage_ids:
            df = history_content_df[history_content_df[HOMEPAGE_ID] == idx]
            homepage_id_content_view_count = (
                df.groupby([CONTENT_ID]).size().to_frame(VIEW_COUNT)
            )
            homepage_id_total_duration_content = history_content_df.groupby([CONTENT_ID]).sum()
            homepage_id_total_duration_content = homepage_id_total_duration_content[[DURATION]]
            final_recommendation_df = merge(
                homepage_id_total_duration_content,
                homepage_id_content_view_count,
                left_index=True,
                right_index=True,
            ).sort_values(by=[DURATION, VIEW_COUNT], ascending=False)
            scaled_features_df = MinMaxScaler(feature_range=(0.1, 1)).fit_transform(
                final_recommendation_df.values
            )
            scaled_features_df = DataFrame(
                scaled_features_df,
                index=final_recommendation_df.index,
                columns=final_recommendation_df.columns,
            ).rename(columns={DURATION: SCALED_DURATION, VIEW_COUNT: SCALED_VIEW_COUNT})
            final_recommendation_df = final_recommendation_df.merge(
                scaled_features_df, how=INNER, left_index=True, right_index=True
            )
            final_recommendation_df[RECOMMENDATION_SCORE] = round(
                (final_recommendation_df[SCALED_DURATION] * 0.7)
                + (final_recommendation_df[SCALED_VIEW_COUNT] * 0.3),
                3,
            )
            Logging.info("Preparing Seasonal Model Dataframe Output")
            final_recommendation_df = final_recommendation_df.sort_values(
                by=RECOMMENDATION_SCORE, ascending=False
            )
            final_recommendation_df = final_recommendation_df.assign(
                rank_score=[i for i in range(len(final_recommendation_df), 0, -1)]
            )
            final_recommendation_df[SCORE] = (
                    final_recommendation_df[RECOMMENDATION_SCORE]
                    + final_recommendation_df[RANK_SCORE]
            )
            final_recommendation_df[[SCORE]] = MinMaxScaler(
                feature_range=(0.1, 1)
            ).fit_transform(final_recommendation_df[[SCORE]])
            final_recommendation_df[SCORE] = final_recommendation_df[SCORE].apply(
                lambda x: round(x, 3)
            )
            final_recommendation_df[CREATED_ON] = datetime.utcnow().isoformat()
            tmp_seasonal_content_df = (
                final_recommendation_df[[SCORE, CREATED_ON]]
                .rename_axis(CONTENT_ID)
                .reset_index()
            )
            seasonal_content_df = concat([seasonal_content_df, tmp_seasonal_content_df], ignore_index=True)
        return seasonal_content_df

    @staticmethod
    @custom_exception()
    def seasonal_algorithm(graph, user_label, homepage_id_wise):
        (
            history_content_df,
            content_homepage_id_mapping,
        ) = SeasonalModel.fetch_historical_data(graph, user_label, homepage_id_wise)

        Logging.info("Calculating Seasonal Score from " + user_label + " UBD data")
        if homepage_id_wise:
            seasonal_content_df = SeasonalModel.homepage_id_wise_scaling(history_content_df)
            seasonal_content_df = merge(
                seasonal_content_df,
                content_homepage_id_mapping,
                on=CONTENT_ID,
                how=LEFT,
            )
            seasonal_content_df = seasonal_content_df[
                ~seasonal_content_df[HOMEPAGE_ID].isnull()
            ].reset_index(drop=True)
            seasonal_content_df = SeasonalModel.filter_by_dedicated_homepage_id(
                seasonal_content_df, user_label
            )
            seasonal_content_df = seasonal_content_df.sort_values(
                by=[HOMEPAGE_ID, SCORE], ascending=[True, False]
            )
        else:
            season_df_view_count = (
                history_content_df.groupby([CONTENT_ID]).size().to_frame(VIEW_COUNT)
            )
            season_df_total_duration_content = history_content_df.groupby([CONTENT_ID]).sum()
            season_df_total_duration_content = season_df_total_duration_content[[DURATION]]
            final_recommendation_df = merge(
                season_df_total_duration_content,
                season_df_view_count,
                left_index=True,
                right_index=True,
            ).sort_values(by=[DURATION, VIEW_COUNT], ascending=False)
            scaled_features_df = MinMaxScaler(feature_range=(0.1, 1)).fit_transform(
                final_recommendation_df.values
            )
            scaled_features_df = DataFrame(
                scaled_features_df,
                index=final_recommendation_df.index,
                columns=final_recommendation_df.columns,
            ).rename(columns={DURATION: SCALED_DURATION, VIEW_COUNT: SCALED_VIEW_COUNT})
            final_recommendation_df = final_recommendation_df.merge(
                scaled_features_df, how=INNER, left_index=True, right_index=True
            )
            final_recommendation_df[RECOMMENDATION_SCORE] = round(
                (final_recommendation_df[SCALED_DURATION] * 0.7)
                + (final_recommendation_df[SCALED_VIEW_COUNT] * 0.3),
                3,
            )
            Logging.info("Preparing Seasonal Model Dataframe Output")
            final_recommendation_df = final_recommendation_df.sort_values(
                by=RECOMMENDATION_SCORE, ascending=False
            )
            final_recommendation_df = final_recommendation_df.assign(
                rank_score=[i for i in range(len(final_recommendation_df), 0, -1)]
            )
            final_recommendation_df[SCORE] = (
                    final_recommendation_df[RECOMMENDATION_SCORE]
                    + final_recommendation_df[RANK_SCORE]
            )
            final_recommendation_df[[SCORE]] = MinMaxScaler(
                feature_range=(0.1, 1)
            ).fit_transform(final_recommendation_df[[SCORE]])
            final_recommendation_df[SCORE] = final_recommendation_df[SCORE].apply(
                lambda x: round(x, 3)
            )
            final_recommendation_df[CREATED_ON] = datetime.utcnow().isoformat()
            seasonal_content_df = (
                final_recommendation_df[[SCORE, CREATED_ON]]
                .rename_axis(CONTENT_ID)
                .reset_index()
            )
            seasonal_content_df = merge(
                seasonal_content_df,
                content_homepage_id_mapping,
                on=CONTENT_ID,
                how=LEFT,
            )
            seasonal_content_df = seasonal_content_df[
                ~seasonal_content_df[HOMEPAGE_ID].isnull()
            ].reset_index(drop=True)
            seasonal_content_df = SeasonalModel.filter_by_dedicated_homepage_id(
                seasonal_content_df, user_label
            )
            seasonal_content_df = seasonal_content_df.sort_values(
                by=SCORE, ascending=False
            )
        seasonal_content_df[REC_TYPE] = SEASONAL_MODULE_NAME.upper()
        return seasonal_content_df

    @staticmethod
    @custom_exception()
    def get_seasonal_content_all_content(graph):
        SeasonalModel.get_data()
        list_of_key_prefix = []
        homepage_id_wise = False
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Seasonal Content based on All Content")
        start = time.time()
        for user_label in list_user_label:
            Logging.info(
                "Start generating seasonal content based on All Content for "
                + user_label
                + " users"
            )
            output_df = SeasonalModel.seasonal_algorithm(
                graph, user_label, homepage_id_wise
            )
            output_df = output_df.drop_duplicates(subset=[CONTENT_ID], keep="first", ignore_index=True)
            Logging.info("Success Preparing Seasonal Dataframe Output")
            key_prefix = (
                    SERVICE_NAME
                    + ":"
                    + SEASONAL_MODULE_NAME.lower()
                    + ":"
                    + user_label
                    + ":"
                    + ALL_CONTENT_BASED
            )
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        seasonal_pay_tv_all_content, seasonal_no_pay_tv_all_content = list_output_df

        Logging.info("Preparing the output to JSON Schema")
        # Convert dataframe output to json output
        seasonal_pay_tv_all_content_dict = SeasonalModel.get_dict_format_output(
            seasonal_pay_tv_all_content, list_of_key_prefix[0], homepage_id_wise
        )
        seasonal_no_pay_tv_all_content_dict = SeasonalModel.get_dict_format_output(
            seasonal_no_pay_tv_all_content, list_of_key_prefix[1], homepage_id_wise
        )
        end = time.time()
        duration = end - start
        Logging.info(
            "Duration for Get Seasonal Content based on All Content is "
            + str(duration)
            + " seconds"
        )
        SeasonalModel.clear_data()
        return seasonal_pay_tv_all_content_dict, seasonal_no_pay_tv_all_content_dict

    @staticmethod
    @custom_exception()
    def get_seasonal_content_homepage_id(graph):
        SeasonalModel.clear_data()
        SeasonalModel.get_data()
        list_of_key_prefix = []
        homepage_id_wise = True
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Seasonal Content based on Homepage_id")
        start = time.time()
        for user_label in list_user_label:
            Logging.info(
                "Start generating seasonal content based on Homepage_id for "
                + user_label
                + " users"
            )
            output_df = SeasonalModel.seasonal_algorithm(
                graph, user_label, homepage_id_wise
            ).reset_index(drop=True)
            Logging.info("Success Preparing Seasonal Dataframe Output")
            key_prefix = (
                    SERVICE_NAME
                    + ":"
                    + SEASONAL_MODULE_NAME.lower()
                    + ":"
                    + user_label
                    + ":"
                    + HOMEPAGE_ID_BASED
            )
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        seasonal_pay_tv_homepage_id, seasonal_no_pay_tv_homepage_id = list_output_df

        Logging.info("Preparing the output to JSON Schema")
        # Convert dataframe output to json output
        seasonal_pay_tv_homepage_id_dict = SeasonalModel.get_dict_format_output(
            seasonal_pay_tv_homepage_id, list_of_key_prefix[0], homepage_id_wise
        )
        seasonal_no_pay_tv_homepage_id_dict = SeasonalModel.get_dict_format_output(
            seasonal_no_pay_tv_homepage_id, list_of_key_prefix[1], homepage_id_wise
        )

        end = time.time()
        duration = end - start
        Logging.info(
            "Duration for Get Seasonal Content based on Homepage_id is "
            + str(duration)
            + " seconds"
        )
        return seasonal_pay_tv_homepage_id_dict, seasonal_no_pay_tv_homepage_id_dict
